/*
Copyright 2011 Daniel Morton. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice, this list
      of conditions and the following disclaimer in the documentation and/or other materials
      provided with the distribution.

THIS SOFTWARE IS PROVIDED BY Daniel Morton ''AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those of the
authors and should not be interpreted as representing official policies, either expressed
or implied, of Daniel Morton.
*/

package ca.quadrilateral.wallgenerator;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;
import java.awt.image.ConvolveOp;
import java.awt.image.Kernel;
import java.util.HashMap;
import java.util.Map;

import javax.media.jai.PerspectiveTransform;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.mortennobel.imagescaling.AdvancedResizeOp;
import com.mortennobel.imagescaling.AdvancedResizeOp.UnsharpenMask;
import com.mortennobel.imagescaling.MultiStepRescaleOp;
import com.mortennobel.imagescaling.ResampleOp;

public class Generator {
	private static final Logger log = LoggerFactory.getLogger(Generator.class);
	private static final float ZERO_PERCENT_THRESHOLD = 0.009f;
	
    private final Config config;
    private final PostProcessingConfig postProcessingConfig;
    
    public Generator(Config config, PostProcessingConfig postProcessingConfig) {
        this.config = config;
        this.postProcessingConfig = postProcessingConfig;
    }

    public Map<String, BufferedImage> generate(BufferedImage sourceImage) {
        final Map<String, BufferedImage> resultMap = new HashMap<String, BufferedImage>();

        final BufferedImage baseImage = scaleImage(sourceImage, "CloseFront");

        resultMap.put("CloseFront", baseImage);

        final Dimension closeFrontSideSize = config.getWallSize("CloseFrontSide");
        final Dimension nearFrontSideSize = config.getWallSize("NearFrontSide");

        final BufferedImage baseImageRightHalf = baseImage.getSubimage(baseImage.getWidth() / 2, 0, baseImage.getWidth() / 2, baseImage.getHeight());
        final BufferedImage closeFrontRight = baseImage.getSubimage(baseImage.getWidth() - closeFrontSideSize.width, 0, closeFrontSideSize.width, baseImage.getHeight());

        resultMap.put("CloseFrontRight", closeFrontRight);

        final BufferedImage baseImageLeftHalf = baseImage.getSubimage(0, 0, baseImage.getWidth() / 2, baseImage.getHeight());
        final BufferedImage closeFrontLeft = baseImage.getSubimage(0, 0, closeFrontSideSize.width, baseImage.getHeight());

        resultMap.put("CloseFrontLeft", closeFrontLeft);

        BufferedImage nearByFront = scaleImage(baseImage, "NearFront");

        final BufferedImage nearByFrontOriginal = new BufferedImage(nearByFront.getWidth(), nearByFront.getHeight(), BufferedImage.TYPE_INT_ARGB);

        final Graphics2D nearByFrontOriginalGraphics = nearByFrontOriginal.createGraphics();
        nearByFrontOriginalGraphics.drawImage(nearByFront, 0, 0, null);
        nearByFrontOriginalGraphics.dispose();
        
        nearByFront = getDarkenedSourceImage(nearByFront, (float)postProcessingConfig.nearDarkening, (float)postProcessingConfig.nearDarkening, true);
        
        final BufferedImage nearByFrontRight = nearByFront.getSubimage(nearByFront.getWidth() - nearFrontSideSize.width, 0, nearFrontSideSize.width, nearByFront.getHeight());
        final BufferedImage nearByFrontLeft = nearByFront.getSubimage(0, 0, nearFrontSideSize.width, nearByFront.getHeight());

        resultMap.put("NearFront", nearByFront);
        resultMap.put("NearFrontRight", nearByFrontRight);
        resultMap.put("NearFrontLeft", nearByFrontLeft);

        BufferedImage farFront = scaleImage(nearByFront, "FarFront");
        
        final BufferedImage farFrontOriginal = new BufferedImage(farFront.getWidth(), farFront.getHeight(), BufferedImage.TYPE_INT_ARGB);
        
        final Graphics2D farFrontOriginalGraphics = farFrontOriginal.createGraphics();
        farFrontOriginalGraphics.drawImage(farFront, 0, 0, null);
        farFrontOriginalGraphics.dispose();
                
        farFront = getDarkenedSourceImage(farFront, (float)postProcessingConfig.farDarkening, (float)postProcessingConfig.farDarkening, true);
        
        
        
        resultMap.put("FarFront", farFront);

        resultMap.put("FarFrontSecondary", farFront);

        final BufferedImage closeLeft = generateTransformedImage(baseImageRightHalf, "CloseSide", "CloseLeft");
        resultMap.put("CloseLeft", closeLeft);

        final BufferedImage closeRight = generateTransformedImage(baseImageLeftHalf, "CloseSide", "CloseRight");
        resultMap.put("CloseRight", closeRight);

        final BufferedImage nearLeft = generateTransformedImage(getDarkenedSourceImage(baseImage, 0f, (float)postProcessingConfig.nearDarkening, true), "NearSide", "NearLeft");
        resultMap.put("NearLeft", nearLeft);

        final BufferedImage nearRight = generateTransformedImage(getDarkenedSourceImage(baseImage, 0f, (float)postProcessingConfig.nearDarkening, false), "NearSide", "NearRight");
        resultMap.put("NearRight", nearRight);

        final BufferedImage farLeft = generateTransformedImage(getDarkenedSourceImage(nearByFrontOriginal, (float)postProcessingConfig.nearDarkening, (float)postProcessingConfig.farDarkening, true), "FarSide", "FarLeft");
        resultMap.put("FarLeft", farLeft);

        final BufferedImage farRight = generateTransformedImage(getDarkenedSourceImage(nearByFrontOriginal, (float)postProcessingConfig.nearDarkening, (float)postProcessingConfig.farDarkening, false), "FarSide", "FarRight");
        resultMap.put("FarRight", farRight);

        return resultMap;
    }
    
    
    
    private BufferedImage getDarkenedSourceImage(BufferedImage sourceImage, float startAlpha, float endAlpha, boolean leftToRight) {
    	if (startAlpha > ZERO_PERCENT_THRESHOLD || endAlpha > ZERO_PERCENT_THRESHOLD) {
    		final BufferedImage blackSource = new BufferedImage(1, sourceImage.getHeight(), BufferedImage.TYPE_INT_ARGB);
    		final BufferedImage modifiedSourceImage = new BufferedImage(sourceImage.getWidth(), sourceImage.getHeight(), BufferedImage.TYPE_INT_ARGB);
    		
    		final Graphics2D blackSourceGraphics = blackSource.createGraphics();
    		blackSourceGraphics.setColor(Color.BLACK);
    		blackSourceGraphics.fillRect(0, 0, 1, sourceImage.getHeight());
    		blackSourceGraphics.dispose();
    
    		final Graphics2D graphics = modifiedSourceImage.createGraphics();
    		graphics.drawImage(sourceImage, 0, 0, null);
    		
    		final float delta = (endAlpha - startAlpha) / sourceImage.getWidth();
    		
    		if (leftToRight) {
    			for(int i = 0; i < sourceImage.getWidth(); i++) {
    				final float alpha = startAlpha + (delta * i);
    				graphics.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, alpha));
    				graphics.drawImage(blackSource, i, 0, null);
    			}
    		} else {
    			for(int i = sourceImage.getWidth() - 1; i >= 0; i--) {
    				final float alpha = startAlpha + (delta * ((sourceImage.getWidth() - 1) - i));
    				graphics.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER, alpha));
    				graphics.drawImage(blackSource, i, 0, null);    				
    			}
    		}    		    		    		
    		graphics.dispose();
    		
    		for(int x = 0; x < sourceImage.getWidth(); x++) {
    			for(int y = 0; y < sourceImage.getHeight(); y++) {
    				int sourceAlphaValue = (sourceImage.getRGB(x, y) & 0xFF000000);
    				int targetPixelColor = (modifiedSourceImage.getRGB(x, y) & 0x00FFFFFF);
    				modifiedSourceImage.setRGB(x, y, (targetPixelColor | sourceAlphaValue));
    			}
    		}
    		
    		return modifiedSourceImage;
    	} else {
    		return sourceImage;
    	}
    }

    private BufferedImage generateTransformedImage(BufferedImage sourceImage, String wallSizeKey, String transformKey) {
        final PerspectiveTransform transform = config.getTransform(transformKey);

        final Dimension wallSize = config.getWallSize(wallSizeKey);

        final BufferedImage resultImage = new BufferedImage(wallSize.width, wallSize.height, BufferedImage.TYPE_INT_ARGB);

        applyTransform(sourceImage, resultImage, transform);
        return resultImage;
    }

    private AdvancedResizeOp getImageScaler(int targetWidth, int targetHeight) {
    	AdvancedResizeOp result = null;
    	if (postProcessingConfig.scalingMethod.getClass().equals(String.class)) {
    		result = new ResampleOp(targetWidth, targetHeight);
    	} else {
    		result = new MultiStepRescaleOp(targetWidth, targetHeight, postProcessingConfig.scalingMethod);
    	}
    	log.info("Using Image Scaler: " + result);
    	return result;
    }	
    
    private BufferedImage scaleImage(BufferedImage sourceImage, String configKey) {
        final Dimension wallSize = config.getWallSize(configKey);
        
        BufferedImage scaledImage = null;
        
        final int targetWidth = wallSize.width;
        final int targetHeight = wallSize.height;
        
        if (sourceImage.getHeight() == targetHeight && sourceImage.getWidth() == targetWidth) {
        	log.info("No scaling needed for wall: " + configKey);
        	return sourceImage;
        }	
        
        log.info("Scaling Image for wall: " + configKey);

        scaledImage = getImageScaler(targetWidth, targetHeight).filter(sourceImage, null);
       
        return postProcess(scaledImage);
    }
    
    private BufferedImage postProcess(BufferedImage image) {
    	BufferedImage processedImage = image;
    	if (postProcessingConfig.softening == null || postProcessingConfig.unsharpenMask == UnsharpenMask.None) {
    		processedImage = unsharpMask(processedImage, postProcessingConfig.unsharpenMask);
    		processedImage = soften(processedImage, postProcessingConfig.softening);
    	} else {
    		if (postProcessingConfig.postProcessingOrder == PostProcessingConfig.SOFTEN_BEFORE_UNSHARPEN) {
        		processedImage = soften(processedImage, postProcessingConfig.softening);    			
        		processedImage = unsharpMask(processedImage, postProcessingConfig.unsharpenMask);    			
    		} else if (postProcessingConfig.postProcessingOrder == PostProcessingConfig.UNSHARPEN_BEFORE_SOFTEN) {
        		processedImage = unsharpMask(processedImage, postProcessingConfig.unsharpenMask);
        		processedImage = soften(processedImage, postProcessingConfig.softening);    			
    		}
    	}
    	
    	return processedImage;
    }
    
    private BufferedImage unsharpMask(BufferedImage sourceImage, UnsharpenMask maskValue) {
    	if (maskValue == UnsharpenMask.None) {
    		return sourceImage;
    	}
    	log.info("Applying Unsharpen Mask: " + maskValue);
    	ResampleOp resampleOp = new ResampleOp(sourceImage.getWidth(), sourceImage.getHeight());
    	resampleOp.setUnsharpenMask(maskValue);
    	return resampleOp.filter(sourceImage, null);
    }
    
    
    
    public BufferedImage soften(BufferedImage sourceImage, Float softenFactor) {    	
        if (softenFactor == null || softenFactor <= ZERO_PERCENT_THRESHOLD)
            return sourceImage;
        else {
        	log.info("Applying Softening...");
            float[] softenArray = {0, softenFactor, 0, softenFactor, 1-(softenFactor*4), softenFactor, 0, softenFactor, 0};
            Kernel kernel = new Kernel(3, 3, softenArray);
            ConvolveOp cOp = new ConvolveOp(kernel, ConvolveOp.EDGE_NO_OP, null);
            return cOp.filter(sourceImage, null);
        }
    }
    
/*    
	public BufferedImage getTrilinearScaledInstance(BufferedImage img, int targetWidth,
			int targetHeight) {
		// REMIND: This only works for opaque images...

		// Use multi-step technique: start with original size, then
		// scale down in multiple passes with drawImage()
		// until the target size is reached
		int iw = img.getWidth();
		int ih = img.getHeight();

		Object hint = RenderingHints.VALUE_INTERPOLATION_BILINEAR;
//		int type = (img.getTransparency() == Transparency.OPAQUE) ? BufferedImage.TYPE_INT_RGB
//				: BufferedImage.TYPE_INT_ARGB;

		int type = BufferedImage.TYPE_INT_ARGB;
		
		// First get down to no more than 2x in W & H
		while (iw > targetWidth * 2 || ih > targetHeight * 2) {
			iw = (iw > targetWidth * 2) ? iw / 2 : iw;
			ih = (ih > targetHeight * 2) ? ih / 2 : ih;
			img = scaleImage(img, type, hint, iw, ih);
		}

		// REMIND: Conservative approach:
		// first get W right, then worry about H

		// If still too wide - do a horizontal trilinear blend
		// of img and a half-width img
		if (iw > targetWidth) {
			int iw2 = iw / 2;
			BufferedImage img2 = scaleImage(img, type, hint, iw2, ih);
			if (iw2 < targetWidth) {
				img = scaleImage(img, type, hint, targetWidth, ih);
				img2 = scaleImage(img2, type, hint, targetWidth, ih);
				interp(img2, img, iw - targetWidth, targetWidth - iw2);
			}
			img = img2;
			iw = targetWidth;
		}
		// iw should now be targetWidth or smaller

		// If still too tall - do a vertical trilinear blend
		// of img and a half-height img
		if (ih > targetHeight) {
			int ih2 = ih / 2;
			BufferedImage img2 = scaleImage(img, type, hint, iw, ih2);
			if (ih2 < targetHeight) {
				img = scaleImage(img, type, hint, iw, targetHeight);
				img2 = scaleImage(img2, type, hint, iw, targetHeight);
				interp(img2, img, ih - targetHeight, targetHeight - ih2);
			}
			img = img2;
			ih = targetHeight;
		}
		// ih should now be targetHeight or smaller

		// If we are too small, then it was probably because one of
		// the dimensions was too small from the start.
		if (iw < targetWidth && ih < targetHeight) {
			img = scaleImage(img, type, hint, targetWidth, targetHeight);
		}

		return img;
	}

	private BufferedImage scaleImage(BufferedImage orig, int type, Object hint,
			int w, int h) {
		BufferedImage tmp = new BufferedImage(w, h, type);
		Graphics2D g2 = tmp.createGraphics();
		g2.setRenderingHint(RenderingHints.KEY_INTERPOLATION, hint);
		g2.drawImage(orig, 0, 0, w, h, null);
		g2.dispose();
		return tmp;
	}

	private void interp(BufferedImage img1, BufferedImage img2, int weight1,
			int weight2) {
		float alpha = weight1;
		alpha /= (weight1 + weight2);
		Graphics2D g2 = img1.createGraphics();
		g2.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER,
				alpha));
		g2.drawImage(img2, 0, 0, null);
		g2.dispose();
	}
    */

    private void applyTransform(BufferedImage sourceImage, BufferedImage destinationImage, PerspectiveTransform transform) {
        for(int x = 0; x < destinationImage.getWidth(); x++) {
            for(int y = 0; y < destinationImage.getHeight(); y++) {
                try {
                    final Point2D sourcePoint = new Point2D.Double(x, y);
                    final Point2D destPoint = new Point2D.Double();
                    transform.inverseTransform(sourcePoint, destPoint);

                    int destPointX = (int)destPoint.getX();
                    int destPointY = (int)destPoint.getY();

                    if (destPointX < 0 ||
                            destPointY < 0 ||
                            destPointX >= sourceImage.getWidth() ||
                            destPointY >= sourceImage.getHeight()) {
                        continue;
                    }

                    final int color = sourceImage.getRGB(destPointX, destPointY);
                    destinationImage.setRGB((int)sourcePoint.getX(), (int)sourcePoint.getY(), color);
                } catch (Exception e) {
                    throw new RuntimeException("Error transforming image!", e);
                }
            }
        }
    }
}
